#ifndef __berstream_h
#define __berstream_h

#include <iostream>
#include <string>
#include <sstream>
#include <stack>
#include <exception>

/* Tda pro pedvn vjimek. Je kompatibiln se standardn
 * tdou std::exception. */ 
class BER_Error : public std::exception {
public:  
  enum Code { NOERROR = 0, NOMEM, RANGE, BADITERATOR, BADOPERATOR, NOSEQ,
    UNKNOWN, ERROR_COUNT };
  BER_Error(Code _id = NOERROR) throw();
  virtual const char *what() const throw();
  virtual int ID() const throw() { return id; }
  virtual ~BER_Error() throw();
private:
  Code id;
  static char *messages[];
};

/* tda pro uchovvn znak v poli */
class char_buffer {
public:
  /* jej itertor, co nejuniverzlnj */
  class iterator {
  public:
    iterator() throw();
    iterator(unsigned char *_data_begin, unsigned char *_data_end,
      unsigned char *_position) throw(BER_Error);
    virtual ~iterator() throw();

    /* porovnvac opertory */
    iterator& operator=(const iterator& it) throw();
    bool operator==(const iterator &it) const throw();
    bool operator!=(const iterator &it) const throw() { return !(*this==it); }
    bool operator<(const iterator &it) const throw(BER_Error);
    bool operator>=(const iterator &it) const throw(BER_Error) {
      return !(*this < it);
    }
    bool operator>(const iterator &it) const throw(BER_Error) {
      return *this >= it && *this != it;
    }
    bool operator<=(const iterator &it) const throw(BER_Error) {
      return *this < it || *this == it;
    }

    /* opertory posunu */
    const iterator& operator++() throw(BER_Error); /* prefix */
    const iterator operator++(int) throw(BER_Error); /* postfix */
    const iterator& operator--() throw(BER_Error); /* prefix */
    const iterator operator--(int) throw(BER_Error); /* postfix */
    
    /* dereferencujc opertor */
    unsigned char operator*() const throw(BER_Error);
  private:
    unsigned char *data_begin;
    unsigned char *data_end;
    unsigned char *position;
  };

  /* konstruktor s nastavitelnou velikost kousk pamti, kter se budou
   * v ppad poteby pidvat */
  char_buffer(int _chunk_size = 32768) throw();
  /* koprovac konstruktor & spol. */
  void copy(const char_buffer& cb) throw(BER_Error);
  char_buffer(const char_buffer& cb) throw(BER_Error);
  char_buffer& operator=(const char_buffer& cb) throw(BER_Error);
  /* destruktor */
  virtual ~char_buffer() throw();
  /* funkce pro ukldn znaku do pole */
  virtual void put(unsigned char c) throw(BER_Error);
  /* funkce dvajc itertor ukazujc na zatek a konec pole */
  iterator begin() throw(BER_Error);
  iterator end() throw(BER_Error);
  /* pomocn funkce */
  unsigned long int get_buffer_size() const { return buffer_size; }
  unsigned long int get_data_size() const { return data_ptr; }
protected:  
  /* data */
  unsigned char *data;
  /* ukazatel na konec dat */
  unsigned long int data_ptr;
private:  
  /* velikost pole */
  unsigned long int buffer_size;
  /* velikost kousku na pidn */
  int chunk_size;
  /* pidvac funkce */
  void reallocate() throw(BER_Error);
};

/* ablonov verze pedchoz tdy, umouje pouvat
 * i jin celoseln typy, ne je char */
template<class Type> class T_buffer : public char_buffer {
public:
  /* podd vechno po pedchoz td, ale itertor bude mt jin
   * derefenn opertor, takov, jak je poadovn v zadn */
  class iterator : public char_buffer::iterator {
  public:
    iterator() throw() : char_buffer::iterator() { }
    iterator(unsigned char *_data_begin, unsigned char *_data_end,
      unsigned char *_position) throw(BER_Error) :
      char_buffer::iterator(_data_begin, _data_end, _position) { }
    virtual ~iterator() throw() { }

    /* obecn dereferencujc opertor */
    Type operator*() const throw(BER_Error);
  };

  /* funkce dvajc itertor ukazujc na zatek a konec pole */
  iterator begin() throw(BER_Error) {
    return typename T_buffer<Type>::iterator(data,
      &data[data_ptr], data);
  }
  iterator end() throw(BER_Error) {
    return typename T_buffer<Type>::iterator(data,
      &data[data_ptr], &data[data_ptr]);
  }
};

template<class Type>
Type T_buffer<Type>::iterator::operator*()
  const throw(BER_Error) {
  return static_cast<Type>(char_buffer::iterator::operator*());
}

/* ablonov tda buffer podle zadn - jde o pouh pejmenovn */
template<class Type> class buffer : public T_buffer<Type> { };

/* speciln podoba ablony pro typ char podle zadn */
template<> class buffer<char> : public T_buffer<char> {
public:
  class iterator : public T_buffer<char>::iterator {
  public:
    iterator() throw() : T_buffer<char>::iterator() { }
    iterator(unsigned char *_data_begin, unsigned char *_data_end,
      unsigned char *_position) throw(BER_Error) :
      T_buffer<char>::iterator(_data_begin, _data_end, _position) { }
    virtual ~iterator() throw() { }

    /* speciln dereferencujc opertor */
    std::string operator*() const throw(BER_Error) {
      std::ostringstream s;
      s.width(2);
      s.fill('0');
      s << std::hex << static_cast<unsigned int>(static_cast<unsigned char>(
	T_buffer<char>::iterator::operator*()));
      return s.str();
    }
  };
  
  /* funkce dvajc itertor ukazujc na zatek a konec pole */
  iterator begin() throw(BER_Error) {
    return buffer<char>::iterator(data,
      &data[data_ptr], data);
  }
  iterator end() throw(BER_Error) {
    return buffer<char>::iterator(data,
      &data[data_ptr], &data[data_ptr]);
  }
};

/* obecn proud */
class berstream {
public:
  enum Type {
    UNDEFINED=0, BOOL=1, INT=2, SEQ=16+0x20, STRING=22
  };
  berstream() throw() : b(&b0) { }; 
  berstream(const char_buffer &in) throw(BER_Error) : b0(in), b(&b0) { }
  virtual ~berstream() throw() { }
  void writelength(unsigned long int length) throw(BER_Error);
  unsigned long int readlength(char_buffer::iterator &i) throw(BER_Error);
protected:
  /* zkladn pole */
  char_buffer b0;
  /* odkaz na pole pro data */
  char_buffer *b;
};

/* kdovac proud */
class oberstream : public berstream {
public:
 /* konstruktory a destruktory */
  oberstream() throw() : berstream() { }
  virtual ~oberstream() throw() { }
  /* opertory vstupu: */
  oberstream& operator<<(bool value) throw(BER_Error);
  oberstream& operator<<(int value) throw(BER_Error);
  oberstream& operator<<(unsigned int value) throw(BER_Error);
  oberstream& operator<<(long int value) throw(BER_Error);
  oberstream& operator<<(unsigned long int value) throw(BER_Error);
  oberstream& operator<<(const char *str) throw(BER_Error);
  oberstream& operator<<(const std::string& str) throw(BER_Error);
  /* data u jsou nachystna, tato funkce je jen zkopruje do vstupu */
  void encode(char_buffer &out) throw(BER_Error);
  /* sekvence */
  void begin() throw(BER_Error);
  void end() throw(BER_Error);
private:
  /* zsobnk pol sekvenc */
  std::stack<char_buffer *> bufstack;
};

/* dekdovac proud */
class iberstream : public berstream {
public:
  /* konstruktory a destruktory */
  iberstream(const char_buffer &in) throw(BER_Error);
  virtual ~iberstream() throw() { }
  /* opertory vstupu: */
  iberstream& operator>>(bool &value) throw(BER_Error);
  iberstream& operator>>(int &value) throw(BER_Error);
  iberstream& operator>>(unsigned int &value) throw(BER_Error);
  iberstream& operator>>(long int &value) throw(BER_Error);
  iberstream& operator>>(unsigned long int &value) throw(BER_Error);
  iberstream& operator>>(std::string &str) throw(BER_Error);
private:
  char_buffer::iterator i;
  bool check(char_buffer::iterator& i, Type t, int len = -1) throw(BER_Error);
};

#endif
