#include <iostream>
#include <string>
#include "berstream.h"

using namespace std;

int main() { try {
  string str = "Hello, world!";
  cout << "String: " << str << endl;
  cout << "Putting into buffer..." << endl;
  buffer<char> b;

  for (string::iterator si = str.begin(); si != str.end(); ++si) {
    b.put(*si);
  }
  cout << "Replaying buffer: ";
  for (buffer<char>::iterator bi = b.begin(); bi != b.end(); ++bi) {
    cout << *bi << " ";
  }
  cout << endl;

  {
    cout << "Big buffer test (needs a little more than " << sizeof(int)
         << " MB of memory)..." << endl;
    buffer<int> big;    
    for (long int i = 0; i < (1 << 20); i++)
      big.put(i & 0xff);
    long int size = 0;
    bool ok = true;
    cout << "Replaying buffer: ";
    for (buffer<int>::iterator bi = big.begin(); bi != big.end(); ++bi)
    {
      ok = ok && *bi == (size & 0xff);
      size++;
    }
    cout << size << " elements, test " << (ok ? "passed" : "failed") << endl;
  }

  cout << "oberstream test..." << endl;
  oberstream bo;
  cout << "Adding boolean true" << endl;
  bo << true;
  cout << "Adding boolean false" << endl;
  bo << false;
  cout << "Adding generic int -16384" << endl;
  bo << -16384;
  cout << "Adding generic unsigned int 8192" << endl;
  bo << 8192U;
  cout << "Adding long int -16777216" << endl;
  bo << -16777216L;
  cout << "Adding unsigned long int 262144" << endl;
  bo << 262144UL;
  cout << "Adding string length " << str.length() << endl;
  bo << str.length();
  cout << "Adding the string" << endl;
  bo << str.c_str();
  cout << "Adding the string encapsulated in the sequence!" << endl;
  bo.begin();
  bo << str.c_str();
  bo.end();
  cout << "Done." << endl;
  cout << "Copying data into buffer..." << endl;
  buffer<char> bo_buf;
  bo.encode(bo_buf);
  cout << "Printing the buffer content: " << endl;
  for (buffer<char>::iterator bi = bo_buf.begin(); bi != bo_buf.end(); bi++) {
    cout << *bi << " ";
  }
  
  cout << endl;
  cout << "iberstream test..." << endl;
  iberstream bi(bo_buf);
  bool boolean;
  cout << "Getting boolean true: ";
  bi >> boolean;
  cout << boolean << endl;
  cout << "Getting boolean false: ";
  bi >> boolean;
  cout << boolean << endl;
  int integer;
  cout << "Getting generic int: ";
  bi >> integer;
  cout << integer << endl;
  unsigned int uinteger;
  cout << "Getting generic unsigned int: ";
  bi >> uinteger;
  cout << uinteger << endl;
  long int linteger;
  cout << "Getting long int: ";
  bi >> linteger;
  cout << linteger << endl;
  unsigned long int ulinteger;
  cout << "Getting unsigned long int: ";
  bi >> ulinteger;
  cout << ulinteger << endl;
  cout << "Getting length: ";
  unsigned long int length;
  bi >> length;
  cout << length << endl;
  string str2;
  cout << "Getting the string: ";
  bi >> str2;
  cout << str2 << endl;
  cout << "Getting the string again (this time from sequence :-): ";
  bi >> str2;
  cout << str2 << endl;

  cout << "Advanced oberstream test..." << endl;
  oberstream bo2;
  string bo2_str;
  char tmp[2] = { 0, 0 }; /* workaround for buggy GCC 3.3.3 */
  for (int i = 0; i < 128; i++)
  {
    tmp[0] = i+1;
    bo2_str.append(tmp);
  }
  cout << "Adding " << bo2_str.length() << "-byte string..." << endl;  
  bo2 << bo2_str;
  cout << "Adding sequence 1 2 [3 4 [5 6] 7 8] 9 10..." << endl;
  bo2 << '\x1' << '\x2';
  bo2.begin();
  bo2 << '\x3' << '\x4';
  bo2.begin();
  bo2 << '\x5' << '\x6';
  bo2.end();
  bo2 << '\x7' << '\x8';
  bo2.end();
  bo2 << '\x9' << '\x10';
  buffer<char> bo_buf2;
  bo2.encode(bo_buf2);
  cout << "Printing the buffer content: " << endl;
  for (buffer<char>::iterator bi = bo_buf2.begin(); bi != bo_buf2.end(); bi++) {
    cout << *bi << " ";
  }
  cout << endl;

  cout << "Advanced iberstream test..." << endl;
  iberstream bi2(bo_buf2);
  string bi2_str;
  bi2 >> bi2_str;
  bool bi2_ok = bi2_str.length() == 128;
  if (bi2_ok)
    for (unsigned char i = 0; i < 128; i++)
      bi2_ok = bi2_ok && (unsigned char) bi2_str[i] == i+1;
  cout << endl;
  cout << bi2_str.length() << "-byte string is " << (bi2_ok ? "OK":"incorrect")
       << endl;
 } catch (BER_Error &c) {
  cout << "BER Error: " << c.what() << endl;
} catch (exception &c) {
  cout << "Unhandled exception: " << c.what() << endl;
} catch (...) {
  cout << "Unhandled exception, program terminated." << endl;
} return 0; }
